import json
import os
import shutil

import pytest
from nbformat.notebooknode import NotebookNode

# noinspection PyProtectedMember
from great_expectations.cli.suite import _suite_edit_workflow
from great_expectations.core import ExpectationSuiteValidationResult
from great_expectations.core.batch import BatchRequest
from great_expectations.core.expectation_suite import (
    ExpectationSuite,
    ExpectationSuiteSchema,
)
from great_expectations.core.usage_statistics.anonymizers.types.base import (
    CLISuiteInteractiveFlagCombinations,
)
from great_expectations.data_context.util import file_relative_path
from great_expectations.exceptions import (
    InvalidExpectationConfigurationError,
    SuiteEditNotebookCustomTemplateModuleNotFoundError,
)
from great_expectations.render.renderer.v3.suite_edit_notebook_renderer import (
    SuiteEditNotebookRenderer,
)
from great_expectations.render.renderer_configuration import MetaNotesFormat
from great_expectations.util import get_context
from great_expectations.validator.validator import Validator
from tests.render.test_util import run_notebook


@pytest.fixture
def data_context_v3_custom_notebooks(tmp_path):
    """
    This data_context is *manually* created to have the config we want, vs
    created with DataContext.create()
    """
    project_path = tmp_path
    context_path = os.path.join(project_path, "great_expectations")
    expectations_dir = os.path.join(context_path, "expectations")
    fixture_dir = file_relative_path(__file__, "../../../test_fixtures")
    custom_notebook_assets_dir = os.path.join("v3", "notebook_assets")
    os.makedirs(
        os.path.join(expectations_dir, "my_dag_node"),
        exist_ok=True,
    )
    shutil.copy(
        os.path.join(fixture_dir, "great_expectations_v013_custom_notebooks.yml"),
        str(os.path.join(context_path, "great_expectations.yml")),
    )
    shutil.copy(
        os.path.join(
            fixture_dir,
            "expectation_suites/parameterized_expectation_suite_fixture.json",
        ),
        os.path.join(expectations_dir, "my_dag_node", "default.json"),
    )
    os.makedirs(os.path.join(context_path, "plugins"), exist_ok=True)
    shutil.copytree(
        os.path.join(fixture_dir, custom_notebook_assets_dir),
        str(os.path.join(context_path, "plugins", custom_notebook_assets_dir)),
    )

    return get_context(context_root_dir=context_path)


@pytest.fixture
def data_context_v3_custom_bad_notebooks(tmp_path):
    """
    This data_context is *manually* created to have the config we want, vs
    created with DataContext.create()
    """
    project_path = tmp_path
    context_path = os.path.join(project_path, "great_expectations")
    expectations_dir = os.path.join(context_path, "expectations")
    fixture_dir = file_relative_path(__file__, "../../../test_fixtures")
    custom_notebook_assets_dir = os.path.join("v3", "notebook_assets")
    os.makedirs(
        os.path.join(expectations_dir, "my_dag_node"),
        exist_ok=True,
    )
    shutil.copy(
        os.path.join(fixture_dir, "great_expectations_v013_bad_notebooks.yml"),
        str(os.path.join(context_path, "great_expectations.yml")),
    )
    shutil.copy(
        os.path.join(
            fixture_dir,
            "expectation_suites/parameterized_expectation_suite_fixture.json",
        ),
        os.path.join(expectations_dir, "my_dag_node", "default.json"),
    )
    os.makedirs(os.path.join(context_path, "plugins"), exist_ok=True)
    shutil.copytree(
        os.path.join(fixture_dir, custom_notebook_assets_dir),
        str(os.path.join(context_path, "plugins", custom_notebook_assets_dir)),
    )

    return get_context(context_root_dir=context_path)


@pytest.fixture
def critical_suite_with_citations(empty_data_context) -> ExpectationSuite:
    """
    This hand made fixture has a wide range of expectations, and has a mix of
    metadata including an BasicSuiteBuilderProfiler entry, and citations.
    """
    context = empty_data_context
    schema: ExpectationSuiteSchema = ExpectationSuiteSchema()
    critical_suite: dict = {
        "expectation_suite_name": "critical",
        "meta": {
            "great_expectations_version": "0.13.15+7252.g32fa98e2a.dirty",
            "columns": {
                "npi": {"description": ""},
                "nppes_provider_last_org_name": {"description": ""},
                "nppes_provider_first_name": {"description": ""},
                "nppes_provider_mi": {"description": ""},
                "nppes_credentials": {"description": ""},
                "nppes_provider_gender": {"description": ""},
                "nppes_entity_code": {"description": ""},
                "nppes_provider_street1": {"description": ""},
                "nppes_provider_street2": {"description": ""},
                "nppes_provider_city": {"description": ""},
            },
            "citations": [
                {
                    "citation_date": "2020-02-28T17:34:31.307271",
                    "batch_request": {
                        "datasource_name": "files_datasource",
                        "data_connector_name": "files_data_connector",
                        "data_asset_name": "10k",
                    },
                    "batch_markers": {
                        "ge_load_time": "20200229T013430.655026Z",
                        "pandas_data_fingerprint": "f6037d92eb4c01f976513bc0aec2420d",
                    },
                    "batch_parameters": None,
                    "comment": "BasicSuiteBuilderProfiler added a citation based on the current batch.",
                }
            ],
            "notes": {
                "format": MetaNotesFormat.MARKDOWN,
                "content": [
                    "#### This is an _example_ suite\n\n- This suite was made by quickly glancing at 1000 rows of your data.\n- This is **not a production suite**. It is meant to show examples of expectations.\n- Because this suite was auto-generated using a very basic profiler that does not know your data like you do, many of the expectations may not be meaningful.\n"
                ],
            },
            "BasicSuiteBuilderProfiler": {
                "created_by": "BasicSuiteBuilderProfiler",
                "created_at": 1582838223.843476,
                "batch_request": {
                    "datasource_name": "files_datasource",
                    "data_connector_name": "files_data_connector",
                    "data_asset_name": "10k",
                },
            },
        },
        "expectations": [
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "npi"},
                "meta": {
                    "question": True,
                    "Notes": "There are empty strings that should probably be nulls",
                    "BasicSuiteBuilderProfiler": {"confidence": "very low"},
                },
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "provider_type"},
            },
        ],
        "data_asset_type": "Dataset",
    }
    expectation_suite_dict: dict = schema.loads(json.dumps(critical_suite))
    return ExpectationSuite(**expectation_suite_dict, data_context=context)


@pytest.fixture
def suite_with_multiple_citations(empty_data_context) -> ExpectationSuite:
    """
    A handmade suite with multiple citations each with different batch_request.

    The most recent citation does not have batch_request
    """
    context = empty_data_context
    schema: ExpectationSuiteSchema = ExpectationSuiteSchema()
    critical_suite: dict = {
        "expectation_suite_name": "critical",
        "meta": {
            "great_expectations_version": "0.13.15+7252.g32fa98e2a.dirty",
            "citations": [
                {
                    "citation_date": "2000-01-01T00:00:01.000001",
                    "batch_request": {
                        "datasource_name": "files_datasource",
                        "data_connector_name": "files_data_connector",
                        "data_asset_name": "1k",
                    },
                },
                # This citation is the most recent and has no batch_request
                {
                    "citation_date": "2020-01-01T00:00:01.000001",
                },
                {
                    "citation_date": "1999-01-01T00:00:01.000001",
                    "batch_request": {
                        "datasource_name": "files_datasource",
                        "data_connector_name": "files_data_connector",
                        "data_asset_name": "2k",
                    },
                },
            ],
        },
        "expectations": [
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "npi"},
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "provider_type"},
            },
        ],
        "data_asset_type": "Dataset",
    }
    expectation_suite_dict: dict = schema.loads(json.dumps(critical_suite))
    return ExpectationSuite(**expectation_suite_dict, data_context=context)


@pytest.fixture
def warning_suite(empty_data_context) -> ExpectationSuite:
    """
    This hand made fixture has a wide range of expectations, and has a mix of
    metadata including BasicSuiteBuilderProfiler entries.
    """
    context = empty_data_context
    schema: ExpectationSuiteSchema = ExpectationSuiteSchema()
    warning_suite: dict = {
        "expectation_suite_name": "warning",
        "meta": {
            "great_expectations_version": "0.13.15+7252.g32fa98e2a.dirty",
            "citations": [
                {
                    "citation_date": "2020-02-28T17:34:31.307271",
                    "batch_request": {
                        "datasource_name": "files_datasource",
                        "data_connector_name": "files_data_connector",
                        "data_asset_name": "10k",
                    },
                    "batch_markers": {
                        "ge_load_time": "20200229T013430.655026Z",
                        "pandas_data_fingerprint": "f6037d92eb4c01f976513bc0aec2420d",
                    },
                    "batch_parameters": None,
                    "comment": "BasicSuiteBuilderProfiler added a citation based on the current batch.",
                }
            ],
        },
        "expectations": [
            {
                "expectation_type": "expect_table_row_count_to_be_between",
                "kwargs": {"min_value": 800000, "max_value": 1200000},
            },
            {
                "expectation_type": "expect_table_column_count_to_equal",
                "kwargs": {"value": 71},
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "npi"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "provider_type"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "nppes_provider_last_org_name"},
            },
            {
                "expectation_type": "expect_column_values_to_be_in_set",
                "kwargs": {
                    "column": "nppes_provider_gender",
                    "value_set": ["M", "F", ""],
                },
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "nppes_entity_code"},
            },
            {
                "expectation_type": "expect_column_values_to_be_in_set",
                "kwargs": {"column": "nppes_entity_code", "value_set": ["I", "O"]},
            },
            {
                "expectation_type": "expect_column_kl_divergence_to_be_less_than",
                "kwargs": {
                    "column": "nppes_entity_code",
                    "partition_object": {
                        "values": ["I", "O"],
                        "weights": [0.9431769750233306, 0.056823024976669335],
                    },
                    "threshold": 0.1,
                },
            },
            {
                "expectation_type": "expect_column_values_to_be_in_set",
                "kwargs": {
                    "column": "nppes_provider_state",
                    "value_set": [
                        "AL",
                        "AK",
                        "AZ",
                        "AR",
                        "CA",
                        "CO",
                        "CT",
                        "DE",
                        "FL",
                        "GA",
                        "HI",
                        "ID",
                        "IL",
                        "IN",
                        "IA",
                        "KS",
                        "KY",
                        "LA",
                        "ME",
                        "MD",
                        "MA",
                        "MI",
                        "MN",
                        "MS",
                        "MO",
                        "MT",
                        "NE",
                        "NV",
                        "NH",
                        "NJ",
                        "NM",
                        "NY",
                        "NC",
                        "ND",
                        "OH",
                        "OK",
                        "OR",
                        "PA",
                        "RI",
                        "SC",
                        "SD",
                        "TN",
                        "TX",
                        "UT",
                        "VT",
                        "VA",
                        "WA",
                        "WV",
                        "WI",
                        "WY",
                        "DC",
                        "PR",
                        "AE",
                        "VI",
                    ],
                    "mostly": 0.999,
                },
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "medicare_participation_indicator"},
            },
            {
                "expectation_type": "expect_column_values_to_be_in_set",
                "kwargs": {
                    "column": "medicare_participation_indicator",
                    "value_set": ["Y", "N"],
                },
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "number_of_hcpcs"},
            },
            {
                "expectation_type": "expect_column_values_to_be_between",
                "kwargs": {
                    "column": "number_of_hcpcs",
                    "min_value": 0,
                    "max_value": 500,
                    "mostly": 0.999,
                },
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "total_unique_benes"},
            },
            {
                "expectation_type": "expect_column_values_to_be_between",
                "kwargs": {
                    "column": "total_unique_benes",
                    "min_value": 0,
                    "max_value": 2000,
                    "mostly": 0.95,
                },
            },
            {
                "expectation_type": "expect_column_values_to_be_null",
                "kwargs": {"column": "med_suppress_indicator", "mostly": 0.85},
            },
            {
                "expectation_type": "expect_column_values_to_be_in_set",
                "kwargs": {"column": "med_suppress_indicator", "value_set": ["#", "*"]},
            },
            {
                "expectation_type": "expect_column_values_to_be_between",
                "kwargs": {
                    "column": "beneficiary_average_age",
                    "min_value": 40,
                    "max_value": 90,
                    "mostly": 0.995,
                },
            },
            {
                "expectation_type": "expect_column_kl_divergence_to_be_less_than",
                "kwargs": {
                    "column": "beneficiary_average_age",
                    "partition_object": {
                        "bins": [8, 16.5, 25, 33.5, 42, 50.5, 59, 67.5, 76, 84.5, 93],
                        "weights": [
                            0.00025259576594384474,
                            0.00013318685840675451,
                            0.0009653750909344757,
                            0.0012363414580378728,
                            0.01081660996274442,
                            0.030813927854975127,
                            0.13495227317818748,
                            0.6919590041664524,
                            0.1244213260634741,
                            0.004449359600843578,
                        ],
                    },
                    "threshold": 0.9,
                },
            },
            {
                "expectation_type": "expect_column_values_to_be_between",
                "kwargs": {
                    "column": "total_submitted_chrg_amt",
                    "min_value": 2000,
                    "max_value": 5000000,
                    "mostly": 0.98,
                },
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "kwargs": {"column": "nppes_provider_first_name", "mostly": 0.9},
            },
            {
                "expectation_type": "expect_column_values_to_match_regex",
                "kwargs": {
                    "column": "nppes_provider_zip",
                    "regex": "^\\d*$",
                    "mostly": 0.999,
                },
            },
        ],
        "data_asset_type": "Dataset",
    }
    expectation_suite_dict: dict = schema.loads(json.dumps(warning_suite))
    return ExpectationSuite(**expectation_suite_dict, data_context=context)


def test_render_with_no_column_cells_without_batch_request(
    critical_suite_with_citations, empty_data_context
):
    critical_suite_with_citations.expectations = []
    obs: dict = SuiteEditNotebookRenderer.from_data_context(
        data_context=empty_data_context
    ).render(suite=critical_suite_with_citations)
    assert isinstance(obs, dict)
    expected = {
        "nbformat": 4,
        "nbformat_minor": 5,
        "metadata": {},
        "cells": [
            {
                "id": "smooth-texture",
                "cell_type": "markdown",
                "source": "# Edit Your Expectation Suite\nUse this notebook to recreate and modify your expectation suite:\n\n**Expectation Suite Name**: `critical`\n",
                "metadata": {},
            },
            {
                "id": "genetic-canada",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'import datetime\n\nimport pandas as pd\n\nimport great_expectations as gx\nimport great_expectations.jupyter_ux\nfrom great_expectations.core.expectation_configuration import ExpectationConfiguration\nfrom great_expectations.data_context.types.resource_identifiers import (\n    ExpectationSuiteIdentifier,\n)\nfrom great_expectations.exceptions import DataContextError\n\ncontext = gx.get_context()\n\n\n# Feel free to change the name of your suite here. Renaming this will not remove the other one.\nexpectation_suite_name = "critical"\ntry:\n    suite = context.get_expectation_suite(expectation_suite_name=expectation_suite_name)\n    print(\n        f\'Loaded ExpectationSuite "{suite.expectation_suite_name}" containing {len(suite.expectations)} expectations.\'\n    )\nexcept DataContextError:\n    suite = context.add_expectation_suite(expectation_suite_name=expectation_suite_name)\n    print(f\'Created ExpectationSuite "{suite.expectation_suite_name}".\')',
                "outputs": [],
            },
            {
                "id": "strategic-exhibit",
                "cell_type": "markdown",
                "source": "## Create & Edit Expectations\n\n\nYou are adding Expectation configurations to the suite. Since you selected manual mode, there is no sample batch of data and no validation happens during this process. See our documentation for more info and examples: **[How to create a new Expectation Suite without a sample batch](https://docs.greatexpectations.io/docs/guides/expectations/how_to_create_and_edit_expectations_based_on_domain_knowledge_without_inspecting_data_directly)**.\n\nNote that if you do use interactive mode you may specify a sample batch of data to use when creating your Expectation Suite. You can then use a `validator` to get immediate feedback on your Expectations against your specified sample batch.\n\n\nYou can see all the available expectations in the **[expectation gallery](https://greatexpectations.io/expectations)**.",
                "metadata": {},
            },
            {
                "id": "interracial-integration",
                "cell_type": "markdown",
                "source": "### Table Expectation(s)",
                "metadata": {},
            },
            {
                "id": "original-punch",
                "cell_type": "markdown",
                "source": "No table level expectations are in this suite. Feel free to add some here.\n",
                "metadata": {},
            },
            {
                "id": "occasional-honolulu",
                "cell_type": "markdown",
                "source": "### Column Expectation(s)",
                "metadata": {},
            },
            {
                "id": "removed-failing",
                "cell_type": "markdown",
                "source": "No column level expectations are in this suite. Feel free to add some here.\n",
                "metadata": {},
            },
            {
                "id": "bigger-clone",
                "cell_type": "markdown",
                "source": "## Review & Save Your Expectations\n\nLet's save the expectation suite as a JSON file in the `great_expectations/expectations` directory of your project.\n\nLet's now rebuild your Data Docs, which helps you communicate about your data with both machines and humans.",
                "metadata": {},
            },
            {
                "id": "broad-weekend",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": "print(context.get_expectation_suite(expectation_suite_name=expectation_suite_name))\ncontext.add_or_update_expectation_suite(expectation_suite=suite)\n\nsuite_identifier = ExpectationSuiteIdentifier(expectation_suite_name=expectation_suite_name)\ncontext.build_data_docs(resource_identifiers=[suite_identifier])\ncontext.open_data_docs(resource_identifier=suite_identifier)",
                "outputs": [],
            },
        ],
    }

    del expected["nbformat_minor"]
    del obs["nbformat_minor"]
    for obs_cell, expected_cell in zip(obs["cells"], expected["cells"]):
        obs_cell.pop("id", None)
        expected_cell.pop("id", None)
        assert obs_cell == expected_cell
    assert obs == expected


def test_complex_suite_with_batch_request(warning_suite, empty_data_context):
    batch_request: dict = {
        "datasource_name": "files_datasource",
        "data_connector_name": "files_data_connector",
        "data_asset_name": "1k",
    }
    obs: NotebookNode = SuiteEditNotebookRenderer.from_data_context(
        data_context=empty_data_context
    ).render(suite=warning_suite, batch_request=batch_request)
    assert isinstance(obs, dict)
    expected: dict = {
        "nbformat": 4,
        "nbformat_minor": 5,
        "metadata": {},
        "cells": [
            {
                "id": "configured-swing",
                "cell_type": "markdown",
                "source": "# Edit Your Expectation Suite\nUse this notebook to recreate and modify your expectation suite:\n\n**Expectation Suite Name**: `warning`\n",
                "metadata": {},
            },
            {
                "id": "proof-employee",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'import datetime\n\nimport pandas as pd\n\nimport great_expectations as gx\nimport great_expectations.jupyter_ux\nfrom great_expectations.core.batch import BatchRequest\nfrom great_expectations.checkpoint import SimpleCheckpoint\nfrom great_expectations.exceptions import DataContextError\n\ncontext = gx.get_context()\n\n# Note that if you modify this batch request, you may save the new version as a .json file\n#  to pass in later via the --batch-request option\nbatch_request = {\n    "datasource_name": "files_datasource",\n    "data_connector_name": "files_data_connector",\n    "data_asset_name": "1k",\n}\n\n\n# Feel free to change the name of your suite here. Renaming this will not remove the other one.\nexpectation_suite_name = "warning"\ntry:\n    suite = context.get_expectation_suite(expectation_suite_name=expectation_suite_name)\n    print(\n        f\'Loaded ExpectationSuite "{suite.expectation_suite_name}" containing {len(suite.expectations)} expectations.\'\n    )\nexcept DataContextError:\n    suite = context.add_expectation_suite(expectation_suite_name=expectation_suite_name)\n    print(f\'Created ExpectationSuite "{suite.expectation_suite_name}".\')\n\n\nvalidator = context.get_validator(\n    batch_request=BatchRequest(**batch_request),\n    expectation_suite_name=expectation_suite_name,\n)\ncolumn_names = [f\'"{column_name}"\' for column_name in validator.columns()]\nprint(f"Columns: {\', \'.join(column_names)}.")\nvalidator.head(n_rows=5, fetch_all=False)',
                "outputs": [],
            },
            {
                "id": "banned-television",
                "cell_type": "markdown",
                "source": "## Create & Edit Expectations\n\n\nAdd expectations by calling specific expectation methods on the `validator` object. They all begin with `.expect_` which makes autocompleting easy using tab.\n\nBecause you selected interactive mode, you are now creating or editing an Expectation Suite with validator feedback from the sample batch of data that you specified (see `batch_request`).\n\nNote that if you select manual mode you may still create or edit an Expectation Suite directly, without feedback from the `validator`. See our documentation for more info and examples: [How to create a new Expectation Suite without a sample batch](https://docs.greatexpectations.io/docs/guides/expectations/how_to_create_and_edit_expectations_based_on_domain_knowledge_without_inspecting_data_directly).\n\n\n\nYou can see all the available expectations in the **[expectation gallery](https://greatexpectations.io/expectations)**.",
                "metadata": {},
            },
            {
                "id": "offensive-rings",
                "cell_type": "markdown",
                "source": "### Table Expectation(s)",
                "metadata": {},
            },
            {
                "id": "stopped-roulette",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": "validator.expect_table_row_count_to_be_between(min_value=800000, max_value=1200000)",
                "outputs": [],
            },
            {
                "id": "extra-decision",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": "validator.expect_table_column_count_to_equal(value=71)",
                "outputs": [],
            },
            {
                "id": "offshore-crown",
                "cell_type": "markdown",
                "source": "### Column Expectation(s)",
                "metadata": {},
            },
            {
                "id": "hollywood-harrison",
                "cell_type": "markdown",
                "source": "#### `npi`",
                "metadata": {},
            },
            {
                "id": "sufficient-address",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="npi")',
                "outputs": [],
            },
            {
                "id": "lightweight-tomato",
                "cell_type": "markdown",
                "source": "#### `provider_type`",
                "metadata": {},
            },
            {
                "id": "advanced-chorus",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="provider_type")',
                "outputs": [],
            },
            {
                "id": "otherwise-creation",
                "cell_type": "markdown",
                "source": "#### `nppes_provider_last_org_name`",
                "metadata": {},
            },
            {
                "id": "infrared-exception",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="nppes_provider_last_org_name")',
                "outputs": [],
            },
            {
                "id": "judicial-highland",
                "cell_type": "markdown",
                "source": "#### `nppes_provider_gender`",
                "metadata": {},
            },
            {
                "id": "linear-portrait",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_in_set(\n    column="nppes_provider_gender", value_set=["M", "F", ""]\n)',
                "outputs": [],
            },
            {
                "id": "senior-beatles",
                "cell_type": "markdown",
                "source": "#### `nppes_entity_code`",
                "metadata": {},
            },
            {
                "id": "genetic-vaccine",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="nppes_entity_code")',
                "outputs": [],
            },
            {
                "id": "stable-budget",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_in_set(\n    column="nppes_entity_code", value_set=["I", "O"]\n)',
                "outputs": [],
            },
            {
                "id": "prescribed-discretion",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_kl_divergence_to_be_less_than(\n    column="nppes_entity_code",\n    partition_object={\n        "values": ["I", "O"],\n        "weights": [0.9431769750233306, 0.056823024976669335],\n    },\n    threshold=0.1,\n)',
                "outputs": [],
            },
            {
                "id": "moral-poison",
                "cell_type": "markdown",
                "source": "#### `nppes_provider_state`",
                "metadata": {},
            },
            {
                "id": "solid-duplicate",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_in_set(\n    column="nppes_provider_state",\n    value_set=[\n        "AL",\n        "AK",\n        "AZ",\n        "AR",\n        "CA",\n        "CO",\n        "CT",\n        "DE",\n        "FL",\n        "GA",\n        "HI",\n        "ID",\n        "IL",\n        "IN",\n        "IA",\n        "KS",\n        "KY",\n        "LA",\n        "ME",\n        "MD",\n        "MA",\n        "MI",\n        "MN",\n        "MS",\n        "MO",\n        "MT",\n        "NE",\n        "NV",\n        "NH",\n        "NJ",\n        "NM",\n        "NY",\n        "NC",\n        "ND",\n        "OH",\n        "OK",\n        "OR",\n        "PA",\n        "RI",\n        "SC",\n        "SD",\n        "TN",\n        "TX",\n        "UT",\n        "VT",\n        "VA",\n        "WA",\n        "WV",\n        "WI",\n        "WY",\n        "DC",\n        "PR",\n        "AE",\n        "VI",\n    ],\n    mostly=0.999,\n)',
                "outputs": [],
            },
            {
                "id": "dried-module",
                "cell_type": "markdown",
                "source": "#### `medicare_participation_indicator`",
                "metadata": {},
            },
            {
                "id": "magnetic-inquiry",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="medicare_participation_indicator")',
                "outputs": [],
            },
            {
                "id": "rough-composer",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_in_set(\n    column="medicare_participation_indicator", value_set=["Y", "N"]\n)',
                "outputs": [],
            },
            {
                "id": "external-virginia",
                "cell_type": "markdown",
                "source": "#### `number_of_hcpcs`",
                "metadata": {},
            },
            {
                "id": "improving-interim",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="number_of_hcpcs")',
                "outputs": [],
            },
            {
                "id": "spiritual-victor",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_between(\n    column="number_of_hcpcs", min_value=0, max_value=500, mostly=0.999\n)',
                "outputs": [],
            },
            {
                "id": "elegant-gates",
                "cell_type": "markdown",
                "source": "#### `total_unique_benes`",
                "metadata": {},
            },
            {
                "id": "demonstrated-variety",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="total_unique_benes")',
                "outputs": [],
            },
            {
                "id": "increasing-newton",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_between(\n    column="total_unique_benes", min_value=0, max_value=2000, mostly=0.95\n)',
                "outputs": [],
            },
            {
                "id": "waiting-graham",
                "cell_type": "markdown",
                "source": "#### `med_suppress_indicator`",
                "metadata": {},
            },
            {
                "id": "willing-adrian",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_null(column="med_suppress_indicator", mostly=0.85)',
                "outputs": [],
            },
            {
                "id": "framed-above",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_in_set(\n    column="med_suppress_indicator", value_set=["#", "*"]\n)',
                "outputs": [],
            },
            {
                "id": "fancy-voltage",
                "cell_type": "markdown",
                "source": "#### `beneficiary_average_age`",
                "metadata": {},
            },
            {
                "id": "accomplished-league",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_between(\n    column="beneficiary_average_age", min_value=40, max_value=90, mostly=0.995\n)',
                "outputs": [],
            },
            {
                "id": "bronze-march",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_kl_divergence_to_be_less_than(\n    column="beneficiary_average_age",\n    partition_object={\n        "bins": [8, 16.5, 25, 33.5, 42, 50.5, 59, 67.5, 76, 84.5, 93],\n        "weights": [\n            0.00025259576594384474,\n            0.00013318685840675451,\n            0.0009653750909344757,\n            0.0012363414580378728,\n            0.01081660996274442,\n            0.030813927854975127,\n            0.13495227317818748,\n            0.6919590041664524,\n            0.1244213260634741,\n            0.004449359600843578,\n        ],\n    },\n    threshold=0.9,\n)',
                "outputs": [],
            },
            {
                "id": "interim-liquid",
                "cell_type": "markdown",
                "source": "#### `total_submitted_chrg_amt`",
                "metadata": {},
            },
            {
                "id": "assured-cannon",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_be_between(\n    column="total_submitted_chrg_amt", min_value=2000, max_value=5000000, mostly=0.98\n)',
                "outputs": [],
            },
            {
                "id": "regular-corrections",
                "cell_type": "markdown",
                "source": "#### `nppes_provider_first_name`",
                "metadata": {},
            },
            {
                "id": "sensitive-landing",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(\n    column="nppes_provider_first_name", mostly=0.9\n)',
                "outputs": [],
            },
            {
                "id": "inappropriate-shift",
                "cell_type": "markdown",
                "source": "#### `nppes_provider_zip`",
                "metadata": {},
            },
            {
                "id": "worthy-rainbow",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_match_regex(\n    column="nppes_provider_zip", regex="^\\d*$", mostly=0.999\n)',
                "outputs": [],
            },
            {
                "id": "provincial-termination",
                "cell_type": "markdown",
                "source": "## Review & Save Your Expectations\n\nLet's save the expectation suite as a JSON file in the `great_expectations/expectations` directory of your project.\n\nLet's now rebuild your Data Docs, which helps you communicate about your data with both machines and humans.",
                "metadata": {},
            },
            {
                "id": "exposed-cement",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'print(validator.get_expectation_suite(discard_failed_expectations=False))\nvalidator.save_expectation_suite(discard_failed_expectations=False)\n\ncheckpoint_config = {\n    "class_name": "SimpleCheckpoint",\n    "validations": [\n        {\n            "batch_request": batch_request,\n            "expectation_suite_name": expectation_suite_name\n        }\n    ]\n}\ncheckpoint = SimpleCheckpoint(\n    f"{validator.active_batch_definition.data_asset_name}_{expectation_suite_name}",\n    context,\n    **checkpoint_config\n)\ncheckpoint_result = checkpoint.run()\n\ncontext.build_data_docs()\n\nvalidation_result_identifier = checkpoint_result.list_validation_result_identifiers()[0]\ncontext.open_data_docs(resource_identifier=validation_result_identifier)',
                "outputs": [],
            },
        ],
    }

    del expected["nbformat_minor"]
    del obs["nbformat_minor"]
    for obs_cell, expected_cell in zip(obs["cells"], expected["cells"]):
        obs_cell.pop("id", None)
        expected_cell.pop("id", None)
        assert obs_cell == expected_cell
    assert obs == expected


@pytest.mark.slow  # 9.72s
def test_notebook_execution_with_pandas_backend(
    titanic_v013_multi_datasource_pandas_data_context_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    """
    To set this test up we:

    - create a suite
    - add a few expectations (both table and column level)
    - verify that no validations have happened
    - create the suite edit notebook by hijacking the private cli method

    We then:
    - execute that notebook (Note this will raise various errors like
    CellExecutionError if any cell in the notebook fails
    - create a new context from disk
    - verify that a validation has been run with our expectation suite
    """
    # Since we'll run the notebook, we use a context with no data docs to avoid the renderer's default
    # behavior of building and opening docs, which is not part of this test.
    context = titanic_v013_multi_datasource_pandas_data_context_with_checkpoints_v1_with_empty_store_stats_enabled
    root_dir: str = context.root_directory
    uncommitted_dir: str = os.path.join(root_dir, "uncommitted")
    expectation_suite_name: str = "warning"

    context.add_expectation_suite(expectation_suite_name=expectation_suite_name)
    batch_request: dict = {
        "datasource_name": "my_datasource",
        "data_connector_name": "my_basic_data_connector",
        "data_asset_name": "Titanic_1912",
    }
    validator: Validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite_name=expectation_suite_name,
    )
    validator.expect_table_column_count_to_equal(1)
    validator.expect_table_row_count_to_equal(1313)
    validator.expect_column_values_to_be_in_set("Sex", ["female", "male"])
    validator.save_expectation_suite(discard_failed_expectations=False)

    # Sanity check test setup
    original_suite: ExpectationSuite = context.get_expectation_suite(
        expectation_suite_name=expectation_suite_name
    )
    assert len(original_suite.expectations) == 3
    assert context.list_expectation_suite_names() == [expectation_suite_name]
    assert context.list_datasources() == [
        {
            "name": "my_datasource",
            "class_name": "Datasource",
            "module_name": "great_expectations.datasource",
            "execution_engine": {
                "class_name": "PandasExecutionEngine",
                "module_name": "great_expectations.execution_engine",
            },
            "data_connectors": {
                "my_basic_data_connector": {
                    "module_name": "great_expectations.datasource.data_connector",
                    "base_directory": f"{root_dir}/../data/titanic",
                    "default_regex": {
                        "pattern": "(.*)\\.csv",
                        "group_names": ["data_asset_name"],
                    },
                    "class_name": "InferredAssetFilesystemDataConnector",
                },
                "my_special_data_connector": {
                    "glob_directive": "*.csv",
                    "assets": {
                        "users": {
                            "pattern": "(.+)_(\\d+)_(\\d+)\\.csv",
                            "group_names": ["name", "timestamp", "size"],
                            "class_name": "Asset",
                            "base_directory": f"{root_dir}/../data/titanic",
                            "module_name": "great_expectations.datasource.data_connector.asset",
                        }
                    },
                    "module_name": "great_expectations.datasource.data_connector",
                    "base_directory": f"{root_dir}/../data/titanic",
                    "default_regex": {"pattern": "(.+)\\.csv", "group_names": ["name"]},
                    "class_name": "ConfiguredAssetFilesystemDataConnector",
                },
                "my_other_data_connector": {
                    "glob_directive": "*.csv",
                    "assets": {
                        "users": {
                            "class_name": "Asset",
                            "module_name": "great_expectations.datasource.data_connector.asset",
                        }
                    },
                    "module_name": "great_expectations.datasource.data_connector",
                    "base_directory": f"{root_dir}/../data/titanic",
                    "default_regex": {"pattern": "(.+)\\.csv", "group_names": ["name"]},
                    "class_name": "ConfiguredAssetFilesystemDataConnector",
                },
                "my_runtime_data_connector": {
                    "module_name": "great_expectations.datasource.data_connector",
                    "batch_identifiers": ["pipeline_stage_name", "airflow_run_id"],
                    "class_name": "RuntimeDataConnector",
                },
            },
        },
        {
            "name": "my_additional_datasource",
            "class_name": "Datasource",
            "module_name": "great_expectations.datasource",
            "execution_engine": {
                "module_name": "great_expectations.execution_engine",
                "class_name": "PandasExecutionEngine",
            },
            "data_connectors": {
                "my_additional_data_connector": {
                    "module_name": "great_expectations.datasource.data_connector",
                    "default_regex": {
                        "pattern": "(.*)\\.csv",
                        "group_names": ["data_asset_name"],
                    },
                    "base_directory": f"{root_dir}/../data/titanic",
                    "class_name": "InferredAssetFilesystemDataConnector",
                }
            },
        },
    ]

    assert (
        context.get_validation_result(expectation_suite_name=expectation_suite_name)
        == {}
    )

    # Create notebook
    # do not want to actually send usage_message, since the function call is not the result of actual usage
    _suite_edit_workflow(
        context=context,
        expectation_suite_name=expectation_suite_name,
        profile=False,
        profiler_name=None,
        usage_event="test_notebook_execution",
        interactive_mode=CLISuiteInteractiveFlagCombinations.UNPROMPTED_INTERACTIVE_FALSE_MANUAL_TRUE,
        no_jupyter=True,
        create_if_not_exist=False,
        datasource_name=None,
        batch_request=batch_request,
        additional_batch_request_args=None,
        suppress_usage_message=True,
    )
    edit_notebook_path: str = os.path.join(uncommitted_dir, "edit_warning.ipynb")
    assert os.path.isfile(edit_notebook_path)

    run_notebook(
        notebook_path=edit_notebook_path,
        notebook_dir=uncommitted_dir,
        string_to_be_replaced="context.open_data_docs(resource_identifier=validation_result_identifier)",
        replacement_string="",
    )

    # Assertions about output
    context = get_context(context_root_dir=root_dir)
    validator: Validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite_name=expectation_suite_name,
    )
    obs_validation_result: ExpectationSuiteValidationResult = validator.validate()
    assert obs_validation_result.statistics == {
        "evaluated_expectations": 3,
        "successful_expectations": 2,
        "unsuccessful_expectations": 1,
        "success_percent": 66.66666666666666,
    }
    suite: ExpectationSuite = context.get_expectation_suite(
        expectation_suite_name=expectation_suite_name
    )
    suite["meta"].pop("citations", None)
    assert suite == original_suite


def test_notebook_execution_with_custom_notebooks_wrong_module(
    suite_with_multiple_citations, data_context_v3_custom_bad_notebooks
):
    """
    Test the error message in case of "bad" custom module is clear
    """
    with pytest.raises(
        SuiteEditNotebookCustomTemplateModuleNotFoundError, match=r"invalid\.module"
    ):
        SuiteEditNotebookRenderer.from_data_context(
            data_context=data_context_v3_custom_bad_notebooks
        ).render(suite=suite_with_multiple_citations)


def test_notebook_execution_with_custom_notebooks(
    suite_with_multiple_citations, data_context_v3_custom_notebooks
):
    """
    Test the different parts of the notebooks can be modified
    """
    batch_request: dict = {
        "datasource_name": "files_datasource",
        "data_connector_name": "files_data_connector",
        "data_asset_name": "1k",
    }
    obs: NotebookNode = SuiteEditNotebookRenderer.from_data_context(
        data_context=data_context_v3_custom_notebooks
    ).render(suite=suite_with_multiple_citations, batch_request=batch_request)
    expected: dict = {
        "nbformat": 4,
        "nbformat_minor": 5,
        "metadata": {},
        "cells": [
            {
                "id": "intellectual-inspection",
                "cell_type": "markdown",
                "source": "# Edit Your Expectation Suite\nUse this notebook to recreate and modify your expectation suite:\n\n**Expectation Suite Name**: `critical`",
                "metadata": {},
            },
            {
                "id": "collaborative-transfer",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'import datetime\n\nimport pandas as pd\n\nimport great_expectations as gx\nimport great_expectations.jupyter_ux\nfrom great_expectations.core.batch import BatchRequest\nfrom great_expectations.checkpoint import SimpleCheckpoint\nfrom great_expectations.exceptions import DataContextError\n\ncontext = gx.get_context()\n\nbatch_request = {\n    "datasource_name": "files_datasource",\n    "data_connector_name": "files_data_connector",\n    "data_asset_name": "1k",\n}\n\n\n# Feel free to change the name of your suite here. Renaming this will not remove the other one.\nexpectation_suite_name = "critical"\ntry:\n    suite = context.get_expectation_suite(expectation_suite_name=expectation_suite_name)\n    print(\n        f\'Loaded ExpectationSuite "{suite.expectation_suite_name}" containing {len(suite.expectations)} expectations.\'\n    )\nexcept DataContextError:\n    suite = context.add_expectation_suite(expectation_suite_name=expectation_suite_name)\n    print(f\'Created ExpectationSuite "{suite.expectation_suite_name}".\')\n\n\nvalidator = context.get_validator(\n    batch_request=BatchRequest(**batch_request),\n    expectation_suite_name=expectation_suite_name,\n)\ncolumn_names = [f\'"{column_name}"\' for column_name in validator.columns()]\nprint(f"Columns: {\', \'.join(column_names)}.")\nvalidator.head(n_rows=5, fetch_all=False)',
                "outputs": [],
            },
            {
                "id": "legal-beauty",
                "cell_type": "markdown",
                "source": "## Create & Edit Expectations\n\n\nAdd expectations by calling specific expectation methods on the `validator` object. They all begin with `.expect_` which makes autocompleting easy using tab.\n\nBecause you selected interactive mode, you are now creating or editing an Expectation Suite with validator feedback from the sample batch of data that you specified (see `batch_request`).\n\nNote that if you select manual mode you may still create or edit an Expectation Suite directly, without feedback from the `validator`. See our documentation for more info and examples: [How to create a new Expectation Suite without a sample batch](https://docs.greatexpectations.io/en/latest/guides/how_to_guides/creating_and_editing_expectations/how_to_create_a_new_expectation_suite_without_a_sample_batch.html).\n\n\n\nYou can see all the available expectations in the **[expectation gallery](https://greatexpectations.io/expectations)**.",
                "metadata": {},
            },
            {
                "id": "smoking-bangladesh",
                "cell_type": "markdown",
                "source": "### Table Expectation(s)",
                "metadata": {},
            },
            {
                "id": "irish-storm",
                "cell_type": "markdown",
                "source": "No table level expectations are in this suite. Feel free to add some here.\n\nThey all begin with `validator.expect_table_...`.\n",
                "metadata": {},
            },
            {
                "id": "injured-differential",
                "cell_type": "markdown",
                "source": "### Column Expectation(s)",
                "metadata": {},
            },
            {
                "id": "eleven-football",
                "cell_type": "markdown",
                "source": "#### `npi`",
                "metadata": {},
            },
            {
                "id": "eleven-solid",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="npi")',
                "outputs": [],
            },
            {
                "id": "responsible-coverage",
                "cell_type": "markdown",
                "source": "#### `provider_type`",
                "metadata": {},
            },
            {
                "id": "compressed-indiana",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'validator.expect_column_values_to_not_be_null(column="provider_type")',
                "outputs": [],
            },
            {
                "id": "scheduled-freeware",
                "cell_type": "markdown",
                "source": "## Review & Save Your Expectations\n\nLet's save the expectation suite as a JSON file in the `great_expectations/expectations` directory of your project.\n\nLet's now rebuild your Data Docs, which helps you communicate about your data with both machines and humans.",
                "metadata": {},
            },
            {
                "id": "useful-hearts",
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'print(validator.get_expectation_suite(discard_failed_expectations=False))\nvalidator.save_expectation_suite(discard_failed_expectations=False)\n\ncheckpoint_config = {\n    "class_name": "SimpleCheckpoint",\n    "validations": [\n        {\n            "batch_request": batch_request,\n            "expectation_suite_name": expectation_suite_name\n        }\n    ]\n}\ncheckpoint = SimpleCheckpoint(\n    f"{validator.active_batch_definition.data_asset_name}_{expectation_suite_name}",\n    context,\n    **checkpoint_config\n)\ncheckpoint_result = checkpoint.run()\n\ncontext.build_data_docs()\n\nvalidation_result_identifier = checkpoint_result.list_validation_result_identifiers()[0]\ncontext.open_data_docs(resource_identifier=validation_result_identifier)',
                "outputs": [],
            },
        ],
    }

    del expected["nbformat_minor"]
    del obs["nbformat_minor"]
    for obs_cell, expected_cell in zip(obs["cells"], expected["cells"]):
        obs_cell.pop("id", None)
        expected_cell.pop("id", None)
        assert obs_cell == expected_cell
    assert obs == expected


@pytest.mark.parametrize(
    "row_condition",
    [
        "Sex == 'female'",
        """
        Sex == "female"
        """,
    ],
)
@pytest.mark.slow  # 1.36s
def test_raise_exception_quotes_or_space_with_row_condition(
    row_condition,
    titanic_v013_multi_datasource_pandas_data_context_with_checkpoints_v1_with_empty_store_stats_enabled,
):
    """
    To set this test up we:

    - create a suite
    - add an expectations with a row_condition
    - create the suite edit notebook by hijacking the private cli method
    - chack if the exception is raised
    """
    # Since we'll run the notebook, we use a context with no data docs to avoid the renderer's default
    # behavior of building and opening docs, which is not part of this test.
    context = titanic_v013_multi_datasource_pandas_data_context_with_checkpoints_v1_with_empty_store_stats_enabled
    root_dir: str = context.root_directory
    uncommitted_dir: str = os.path.join(root_dir, "uncommitted")  # noqa: F841
    expectation_suite_name: str = "warning"

    context.add_expectation_suite(expectation_suite_name=expectation_suite_name)
    batch_request: dict = {
        "datasource_name": "my_datasource",
        "data_connector_name": "my_basic_data_connector",
        "data_asset_name": "Titanic_1912",
    }
    validator: Validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite_name=expectation_suite_name,
    )
    with pytest.raises(
        InvalidExpectationConfigurationError,
        match=r".*Do not introduce (?:simple quotes|\\n) in configuration.*",
    ):
        validator.expect_column_values_to_be_in_set(
            column="Sex",
            row_condition=row_condition,
            condition_parser="pandas",
            value_set=["female", "male"],
        )



import copy
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import great_expectations.exceptions as gx_exceptions
from great_expectations.core._docs_decorators import public_api
from great_expectations.core.batch import (
    Batch,
    BatchDefinition,
    BatchMarkers,
    BatchRequest,
    RuntimeBatchRequest,
)
from great_expectations.core.batch_spec import PathBatchSpec  # noqa: TCH001
from great_expectations.data_context.types.base import ConcurrencyConfig  # noqa: TCH001
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.datasource.data_connector import DataConnector  # noqa: TCH001
from great_expectations.execution_engine import ExecutionEngine  # noqa: TCH001

logger = logging.getLogger(__name__)


@public_api
class BaseDatasource:
    """A Datasource is the glue between an `ExecutionEngine` and a `DataConnector`. This class should be considered
    abstract and not directly instantiated; please use `Datasource` instead.

    Args:
        name: the name for the datasource
        execution_engine: the type of compute engine to produce
        data_context_root_directory: Installation directory path (if installed on a filesystem).
        concurrency: Concurrency config used to configure the execution engine.
        id: Identifier specific to this datasource.
    """

    recognized_batch_parameters: set = {"limit"}

    def __init__(
        self,
        name: str,
        execution_engine: Optional[dict] = None,
        data_context_root_directory: Optional[str] = None,
        concurrency: Optional[ConcurrencyConfig] = None,
        id: Optional[str] = None,
    ) -> None:
        """
        Build a new Datasource.

        Args:
            name: the name for the datasource
            execution_engine (ClassConfig): the type of compute engine to produce
            data_context_root_directory: Installation directory path (if installed on a filesystem).
            concurrency: Concurrency config used to configure the execution engine.
            id: Identifier specific to this datasource.
        """
        self._name = name
        self._id = id

        self._data_context_root_directory = data_context_root_directory
        if execution_engine is None:
            raise gx_exceptions.ExecutionEngineError(
                message="No ExecutionEngine configuration provided."
            )

        try:
            self._execution_engine = instantiate_class_from_config(
                config=execution_engine,
                runtime_environment={"concurrency": concurrency},
                config_defaults={"module_name": "great_expectations.execution_engine"},
            )
        except Exception as e:
            raise gx_exceptions.ExecutionEngineError(message=str(e))

        self._datasource_config: dict = {
            "execution_engine": execution_engine,
            "id": id,
            "name": name,
        }

        # Chetan - 20221103 - This attribute is meant to represent the config args used to instantiate the object (before ${VARIABLE} substitution).
        # While downstream logic should override this value, we default to `self._datasource_config` as a backup.
        # This is to be removed once substitution logic is migrated from the context to the individual object level.
        self._raw_config = self._datasource_config

        self._data_connectors: dict = {}

    def get_batch_from_batch_definition(
        self,
        batch_definition: BatchDefinition,
        batch_data: Any = None,
    ) -> Batch:
        """
        Note: this method should *not* be used when getting a Batch from a BatchRequest, since it does not capture BatchRequest metadata.
        """
        if not isinstance(batch_data, type(None)):
            # TODO: <Alex>Abe: Are the comments below still pertinent?  Or can they be deleted?</Alex>
            # NOTE Abe 20201014: Maybe do more careful type checking here?
            # Seems like we should verify that batch_data is compatible with the execution_engine...?
            batch_spec, batch_markers = None, None
        else:
            data_connector: DataConnector = self.data_connectors[
                batch_definition.data_connector_name
            ]
            (
                batch_data,
                batch_spec,
                batch_markers,
            ) = data_connector.get_batch_data_and_metadata(
                batch_definition=batch_definition
            )
        new_batch = Batch(
            data=batch_data,
            batch_request=None,
            batch_definition=batch_definition,
            batch_spec=batch_spec,
            batch_markers=batch_markers,
        )
        return new_batch

    def get_single_batch_from_batch_request(
        self, batch_request: Union[BatchRequest, RuntimeBatchRequest]
    ) -> Batch:
        batch_list: List[Batch] = self.get_batch_list_from_batch_request(batch_request)
        if len(batch_list) != 1:
            raise ValueError(
                f"Got {len(batch_list)} batches instead of a single batch."
            )
        return batch_list[0]

    def get_batch_definition_list_from_batch_request(
        self, batch_request: Union[BatchRequest, RuntimeBatchRequest]
    ) -> List[BatchDefinition]:
        """
        Validates batch request and utilizes the classes'
        Data Connectors' property to get a list of batch definition given a batch request
        Args:
            :param batch_request: A BatchRequest or RuntimeBatchRequest object used to request a batch
            :return: A list of batch definitions
        """
        self._validate_batch_request(batch_request=batch_request)

        data_connector: DataConnector = self.data_connectors[
            batch_request.data_connector_name
        ]
        return data_connector.get_batch_definition_list_from_batch_request(
            batch_request=batch_request
        )

    def get_batch_list_from_batch_request(
        self, batch_request: Union[BatchRequest, RuntimeBatchRequest]
    ) -> List[Batch]:
        """
        Processes batch_request and returns the (possibly empty) list of batch objects.

        Args:
            :batch_request encapsulation of request parameters necessary to identify the (possibly multiple) batches
            :returns possibly empty list of batch objects; each batch object contains a dataset and associated metatada
        """
        self._validate_batch_request(batch_request=batch_request)

        data_connector: DataConnector = self.data_connectors[
            batch_request.data_connector_name
        ]

        batch_definition_list: List[
            BatchDefinition
        ] = data_connector.get_batch_definition_list_from_batch_request(
            batch_request=batch_request
        )

        if isinstance(batch_request, RuntimeBatchRequest):
            # This is a runtime batch_request

            if len(batch_definition_list) != 1:
                raise ValueError(
                    "RuntimeBatchRequests must specify exactly one corresponding BatchDefinition"
                )

            batch_definition = batch_definition_list[0]
            runtime_parameters = batch_request.runtime_parameters

            # noinspection PyArgumentList
            (
                batch_data,
                batch_spec,
                batch_markers,
            ) = data_connector.get_batch_data_and_metadata(  # type: ignore[call-arg]
                batch_definition=batch_definition,
                runtime_parameters=runtime_parameters,
            )

            new_batch = Batch(
                data=batch_data,
                batch_request=batch_request,
                batch_definition=batch_definition,
                batch_spec=batch_spec,
                batch_markers=batch_markers,
            )

            return [new_batch]
        else:
            batches: List[Batch] = []
            for batch_definition in batch_definition_list:
                batch_definition.batch_spec_passthrough = (
                    batch_request.batch_spec_passthrough
                )
                batch_data: Any  # type: ignore[no-redef]
                batch_spec: PathBatchSpec  # type: ignore[no-redef]
                batch_markers: BatchMarkers  # type: ignore[no-redef]
                (
                    batch_data,
                    batch_spec,
                    batch_markers,
                ) = data_connector.get_batch_data_and_metadata(
                    batch_definition=batch_definition
                )
                new_batch = Batch(
                    data=batch_data,
                    batch_request=batch_request,
                    batch_definition=batch_definition,
                    batch_spec=batch_spec,
                    batch_markers=batch_markers,
                )
                batches.append(new_batch)
            return batches

    def _build_data_connector_from_config(
        self,
        name: str,
        config: Dict[str, Any],
    ) -> DataConnector:
        """Build a DataConnector using the provided configuration and return the newly-built DataConnector."""
        new_data_connector: DataConnector = instantiate_class_from_config(
            config=config,
            runtime_environment={
                "name": name,
                "datasource_name": self.name,
                "execution_engine": self.execution_engine,
            },
            config_defaults={
                "module_name": "great_expectations.datasource.data_connector"
            },
        )
        new_data_connector.data_context_root_directory = (
            self._data_context_root_directory  # type: ignore[assignment]
        )

        self.data_connectors[name] = new_data_connector
        return new_data_connector

    @public_api
    def get_available_data_asset_names(
        self, data_connector_names: Optional[Union[list, str]] = None
    ) -> Dict[str, List[str]]:
        """Returns a dictionary of data_asset_names that the specified dataconnector can provide.

        Note that some data_connectors may not be
        capable of describing specific named data assets, and some (such as
        inferred_asset_data_connector) require the user to configure
        data asset names.

        Example return value:
        ```python
        {
          data_connector_name: {
            names: [ data_asset_1, data_asset_2 ... ]
          }
          ...
        }
        ```

        Args:
            data_connector_names: the DataConnector for which to get available data asset names.

        Returns:
            Dictionary consisting of sets of data assets available for the specified data connectors.
        """
        available_data_asset_names: dict = {}
        if data_connector_names is None:
            data_connector_names = self.data_connectors.keys()  # type: ignore[assignment]
        elif isinstance(data_connector_names, str):
            data_connector_names = [data_connector_names]

        for data_connector_name in data_connector_names:  # type: ignore[union-attr]
            data_connector: DataConnector = self.data_connectors[data_connector_name]
            available_data_asset_names[
                data_connector_name
            ] = data_connector.get_available_data_asset_names()

        return available_data_asset_names

    def get_available_data_asset_names_and_types(
        self, data_connector_names: Optional[Union[list, str]] = None
    ) -> Dict[str, List[Tuple[str, str]]]:
        """
        Returns a dictionary of data_asset_names that the specified data
        connector can provide. Note that some data_connectors may not be
        capable of describing specific named data assets, and some (such as
        inferred_asset_data_connector) require the user to configure
        data asset names.

        Returns:
            dictionary consisting of sets of data assets available for the specified data connectors:
            For instance, in a SQL Database the data asset name corresponds to the table or
            view name, and the data asset type is either 'table' or 'view'.
            ::

                {
                  data_connector_name: {
                    names: [ (data_asset_name_1, data_asset_1_type), (data_asset_name_2, data_asset_2_type) ... ]
                  }
                  ...
                }
        """
        # NOTE: Josh 20211001 This feature is only implemented for the InferredAssetSqlDataConnector

        available_data_asset_names_and_types: dict = {}
        if data_connector_names is None:
            data_connector_names = self.data_connectors.keys()  # type: ignore[assignment]
        elif isinstance(data_connector_names, str):
            data_connector_names = [data_connector_names]

        for data_connector_name in data_connector_names:  # type: ignore[union-attr]
            data_connector: DataConnector = self.data_connectors[data_connector_name]
            available_data_asset_names_and_types[
                data_connector_name
            ] = data_connector.get_available_data_asset_names_and_types()

        return available_data_asset_names_and_types

    def get_available_batch_definitions(
        self, batch_request: Union[BatchRequest, RuntimeBatchRequest]
    ) -> List[BatchDefinition]:
        self._validate_batch_request(batch_request=batch_request)

        data_connector: DataConnector = self.data_connectors[
            batch_request.data_connector_name
        ]
        batch_definition_list = (
            data_connector.get_batch_definition_list_from_batch_request(
                batch_request=batch_request
            )
        )

        return batch_definition_list

    def self_check(self, pretty_print=True, max_examples=3):
        # Provide visibility into parameters that ExecutionEngine was instantiated with.
        report_object: dict = {"execution_engine": self.execution_engine.config}

        if pretty_print:
            print(
                f"\nExecutionEngine class name: {self.execution_engine.__class__.__name__}"
            )

        if pretty_print:
            print("Data Connectors:")

        data_connector_list = list(self.data_connectors.keys())
        data_connector_list.sort()
        report_object["data_connectors"] = {"count": len(data_connector_list)}

        for data_connector_name in data_connector_list:
            data_connector_obj: DataConnector = self.data_connectors[
                data_connector_name
            ]
            data_connector_return_obj = data_connector_obj.self_check(
                pretty_print=pretty_print, max_examples=max_examples
            )
            report_object["data_connectors"][
                data_connector_name
            ] = data_connector_return_obj

        return report_object

    def _validate_batch_request(
        self, batch_request: Union[BatchRequest, RuntimeBatchRequest]
    ) -> None:
        if not (
            batch_request.datasource_name is None
            or batch_request.datasource_name == self.name
        ):
            raise ValueError(
                f"""datasource_name in BatchRequest: "{batch_request.datasource_name}" does not
                match Datasource name: "{self.name}".
                """
            )

        if batch_request.data_connector_name not in self.data_connectors.keys():
            raise ValueError(
                f"""data_connector_name in BatchRequest: "{batch_request.data_connector_name}" is not configured for DataSource: "{self.name}".
                    """
            )

    @property
    def name(self) -> str:
        """
        Property for datasource name
        """
        return self._name

    @property
    def id(self) -> str:
        """
        Property for datasource id
        """
        return self._id  # type: ignore[return-value]

    @property
    def execution_engine(self) -> ExecutionEngine:
        return self._execution_engine

    @property
    def data_connectors(self) -> dict:
        return self._data_connectors

    @property
    def config(self) -> dict:
        return copy.deepcopy(self._datasource_config)


@public_api
class Datasource(BaseDatasource):
    """A Datasource is the glue between an `ExecutionEngine` and a `DataConnector`.

    Args:
        name: the name for the datasource
        execution_engine: the type of compute engine to produce
        data_connectors: DataConnectors to add to the datasource
        data_context_root_directory: Installation directory path (if installed on a filesystem).
        concurrency: Concurrency config used to configure the execution engine.
        id: Identifier specific to this datasource.
    """

    recognized_batch_parameters: set = {"limit"}

    def __init__(
        self,
        name: str,
        execution_engine: Optional[dict] = None,
        data_connectors: Optional[dict] = None,
        data_context_root_directory: Optional[str] = None,
        concurrency: Optional[ConcurrencyConfig] = None,
        id: Optional[str] = None,
    ) -> None:
        """
        Build a new Datasource with data connectors.

        Args:
            name: the name for the datasource
            execution_engine (ClassConfig): the type of compute engine to produce
            data_connectors: DataConnectors to add to the datasource
            data_context_root_directory: Installation directory path (if installed on a filesystem).
            concurrency: Concurrency config used to configure the execution engine.
            id: Identifier specific to this datasource.
        """
        self._name = name

        super().__init__(
            name=name,
            execution_engine=execution_engine,
            data_context_root_directory=data_context_root_directory,
            concurrency=concurrency,
            id=id,
        )

        if data_connectors is None:
            data_connectors = {}
        self._data_connectors = data_connectors
        self._datasource_config.update(
            {"data_connectors": copy.deepcopy(data_connectors)}
        )
        self._init_data_connectors(data_connector_configs=data_connectors)

    def _init_data_connectors(
        self,
        data_connector_configs: Dict[str, Dict[str, Any]],
    ) -> None:
        for name, config in data_connector_configs.items():
            self._build_data_connector_from_config(
                name=name,
                config=config,
            )



from great_expectations.core.usage_statistics.events import UsageStatsEvents


def test_get_cli_event_name():
    assert (
        UsageStatsEvents.get_cli_event_name("checkpoint", "delete", ["begin"])
        == "cli.checkpoint.delete.begin"
    )


def test_get_cli_begin_and_end_event_names():
    assert UsageStatsEvents.get_cli_begin_and_end_event_names("datasource", "new") == [
        "cli.datasource.new.begin",
        "cli.datasource.new.end",
    ]



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Mapping, Optional, Tuple, Union

from great_expectations.util import get_context

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr

from great_expectations.core._docs_decorators import deprecated_argument
from great_expectations.data_context.data_context.abstract_data_context import (
    AbstractDataContext,
)
from great_expectations.data_context.types.base import (
    DataContextConfig,  # noqa: TCH001
    GXCloudConfig,  # noqa: TCH001
)


@deprecated_argument(argument_name="ge_cloud_mode", version="0.15.37")
@deprecated_argument(argument_name="ge_cloud_config", version="0.15.37")
def BaseDataContext(
    project_config: Union[DataContextConfig, Mapping],
    context_root_dir: Optional[PathStr] = None,
    runtime_environment: Optional[dict] = None,
    cloud_mode: bool = False,
    cloud_config: Optional[GXCloudConfig] = None,
    # Deprecated as of 0.15.37
    ge_cloud_mode: bool = False,
    ge_cloud_config: Optional[GXCloudConfig] = None,
) -> AbstractDataContext:
    """A lightweight wrapper around `get_context()`.

    While this used to be the canonical method of instantiating a DataContext before 0.15.40,
    it is now recommended to use `get_context()`.

    Usage:
        `import BaseDataContext from great_expectations.data_context`

        `my_context = BaseDataContext(<insert_your_parameters>)`

    This class implements most of the functionality of DataContext, with a few exceptions.


    Args:
        project_config: In-memory configuration for Data Context.
        context_root_dir (str or pathlib.Path): Path to directory that contains great_expectations.yml file
        runtime_environment: A dictionary of values can be passed to a DataContext when it is instantiated.
            These values will override both values from the config variables file and
            from environment variables.
        cloud_config: GX Cloud credentials (base URL, access token, and org id)
        cloud_mode: Whether to run GX in Cloud mode (default is None).
            If None, cloud mode is assumed if Cloud credentials are set up. Set to False to override.
        ge_cloud_config: GX Cloud credentials (base URL, access token, and org id)
        ge_cloud_mode: Whether to run GX in Cloud mode (default is None).
            If None, cloud mode is assumed if Cloud credentials are set up. Set to False to override.

    Returns:
        A Data Context. Either a FileDataContext, EphemeralDataContext, or
        CloudDataContext depending on environment and/or
        parameters.

    Raises:
        GXCloudConfigurationError: Cloud mode enabled, but missing configuration.


    --ge-feature-maturity-info--

        id: os_linux
        title: OS - Linux
        icon:
        short_description:
        description:
        how_to_guide_url:
        maturity: Production
        maturity_details:
            api_stability: N/A
            implementation_completeness: N/A
            unit_test_coverage: Complete
            integration_infrastructure_test_coverage: Complete
            documentation_completeness: Complete
            bug_risk: Low

        id: os_macos
        title: OS - MacOS
        icon:
        short_description:
        description:
        how_to_guide_url:
        maturity: Production
        maturity_details:
            api_stability: N/A
            implementation_completeness: N/A
            unit_test_coverage: Complete (local only)
            integration_infrastructure_test_coverage: Complete (local only)
            documentation_completeness: Complete
            bug_risk: Low

        id: os_windows
        title: OS - Windows
        icon:
        short_description:
        description:
        how_to_guide_url:
        maturity: Beta
        maturity_details:
            api_stability: N/A
            implementation_completeness: N/A
            unit_test_coverage: Minimal
            integration_infrastructure_test_coverage: Minimal
            documentation_completeness: Complete
            bug_risk: Moderate
    ------------------------------------------------------------
        id: workflow_create_edit_expectations_cli_scaffold
        title: Create and Edit Expectations - suite scaffold
        icon:
        short_description: Creating a new Expectation Suite using suite scaffold
        description: Creating Expectation Suites through an interactive development loop using suite scaffold
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/creating_and_editing_expectations/how_to_automatically_create_a_new_expectation_suite.html
        maturity: Experimental (expect exciting changes to Profiler capability)
        maturity_details:
            api_stability: N/A
            implementation_completeness: N/A
            unit_test_coverage: N/A
            integration_infrastructure_test_coverage: Partial
            documentation_completeness: Complete
            bug_risk: Low

        id: workflow_create_edit_expectations_cli_edit
        title: Create and Edit Expectations - CLI
        icon:
        short_description: Creating a new Expectation Suite using the CLI
        description: Creating a Expectation Suite great_expectations suite new command
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/creating_and_editing_expectations/how_to_create_a_new_expectation_suite_using_the_cli.html
        maturity: Experimental (expect exciting changes to Profiler and Suite Renderer capability)
        maturity_details:
            api_stability: N/A
            implementation_completeness: N/A
            unit_test_coverage: N/A
            integration_infrastructure_test_coverage: Partial
            documentation_completeness: Complete
            bug_risk: Low

        id: workflow_create_edit_expectations_json_schema
        title: Create and Edit Expectations - Json schema
        icon:
        short_description: Creating a new Expectation Suite from a json schema file
        description: Creating a new Expectation Suite using JsonSchemaProfiler function and json schema file
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/creating_and_editing_expectations/how_to_create_a_suite_from_a_json_schema_file.html
        maturity: Experimental (expect exciting changes to Profiler capability)
        maturity_details:
            api_stability: N/A
            implementation_completeness: N/A
            unit_test_coverage: N/A
            integration_infrastructure_test_coverage: Partial
            documentation_completeness: Complete
            bug_risk: Low

    --ge-feature-maturity-info--
    """
    # Chetan - 20221208 - not formally deprecating these values until a future date
    cloud_mode, cloud_config = _resolve_cloud_args(
        cloud_mode=cloud_mode,
        cloud_config=cloud_config,
        ge_cloud_mode=ge_cloud_mode,
        ge_cloud_config=ge_cloud_config,
    )

    project_data_context_config: DataContextConfig = (
        AbstractDataContext.get_or_create_data_context_config(project_config)
    )

    if context_root_dir is not None:
        context_root_dir = os.path.abspath(context_root_dir)
    # initialize runtime_environment as empty dict if None
    runtime_environment = runtime_environment or {}

    cloud_base_url: Optional[str] = None
    cloud_access_token: Optional[str] = None
    cloud_organization_id: Optional[str] = None
    if cloud_config:
        cloud_base_url = cloud_config.base_url
        cloud_access_token = cloud_config.access_token
        cloud_organization_id = cloud_config.organization_id

    return get_context(
        project_config=project_data_context_config,
        context_root_dir=context_root_dir,
        runtime_environment=runtime_environment,
        cloud_base_url=cloud_base_url,
        cloud_access_token=cloud_access_token,
        cloud_organization_id=cloud_organization_id,
        cloud_mode=cloud_mode,
    )


def _resolve_cloud_args(
    cloud_mode: bool = False,
    cloud_config: Optional[GXCloudConfig] = None,
    # <GX_RENAME> Deprecated as of 0.15.37
    ge_cloud_mode: bool = False,
    ge_cloud_config: Optional[GXCloudConfig] = None,
) -> Tuple[bool, Optional[GXCloudConfig]]:
    cloud_mode = True if cloud_mode or ge_cloud_mode else False
    cloud_config = cloud_config if cloud_config is not None else ge_cloud_config
    return cloud_mode, cloud_config



import datetime
import json
import os
from typing import List, Optional, Tuple, Union

from ruamel.yaml.comments import CommentedMap

import great_expectations.exceptions as gx_exceptions
from great_expectations import DataContext
from great_expectations.checkpoint import Checkpoint, LegacyCheckpoint  # noqa: TCH001
from great_expectations.cli.upgrade_helpers.base_upgrade_helper import BaseUpgradeHelper
from great_expectations.data_context.store.checkpoint_store import CheckpointStore
from great_expectations.data_context.types.base import (
    DataContextConfig,
    DataContextConfigDefaults,
)


class UpgradeHelperV13(BaseUpgradeHelper):
    def __init__(
        self,
        data_context: Optional[DataContext] = None,
        context_root_dir: Optional[str] = None,
        update_version: bool = False,
    ) -> None:
        assert (
            data_context or context_root_dir
        ), "Please provide a data_context object or a context_root_dir."

        self.data_context = data_context or DataContext(
            context_root_dir=context_root_dir
        )

        # noinspection SpellCheckingInspection
        self.upgrade_log = {
            "update_version": update_version,
            "skipped_checkpoint_store_upgrade": False,
            "added_checkpoint_store": {},
            "skipped_checkpoint_config_upgrade": False,
            "skipped_datasources_upgrade": False,
            "skipped_validation_operators_upgrade": False,
        }

        # noinspection SpellCheckingInspection
        self.upgrade_checklist = {
            "automatic": {
                "stores": {},
                "store_names": {},
            },
            "manual": {
                "checkpoints": {},
                "datasources": {},
                "validation_operators": {},
            },
        }

        self._generate_upgrade_checklist()

    def _generate_upgrade_checklist(self) -> None:
        self._process_checkpoint_store_for_checklist()
        self._process_checkpoint_config_for_checklist()
        self._process_datasources_for_checklist()
        self._process_validation_operators_for_checklist()

    def _process_checkpoint_store_for_checklist(self) -> None:
        if CheckpointStore.default_checkpoints_exist(
            directory_path=self.data_context.root_directory
        ):
            config_commented_map: CommentedMap = (
                self.data_context.get_config().commented_map
            )
            checkpoint_store_name: Optional[str] = config_commented_map.get(
                "checkpoint_store_name"
            )
            stores: dict = config_commented_map["stores"]
            if checkpoint_store_name:
                if stores.get(checkpoint_store_name):
                    self.upgrade_log["skipped_checkpoint_store_upgrade"] = True
                else:
                    self.upgrade_checklist["automatic"]["stores"] = {
                        checkpoint_store_name: DataContextConfigDefaults.DEFAULT_STORES.value[
                            DataContextConfigDefaults.DEFAULT_CHECKPOINT_STORE_NAME.value
                        ]
                    }
            else:
                checkpoint_store_name = (
                    DataContextConfigDefaults.DEFAULT_CHECKPOINT_STORE_NAME.value
                )
                self.upgrade_checklist["automatic"]["store_names"][
                    "checkpoint_store_name"
                ] = checkpoint_store_name
                if not stores.get(checkpoint_store_name):
                    self.upgrade_checklist["automatic"]["stores"] = {
                        checkpoint_store_name: DataContextConfigDefaults.DEFAULT_STORES.value[
                            checkpoint_store_name
                        ]
                    }
        else:
            self.upgrade_log["skipped_checkpoint_store_upgrade"] = True

    def _process_checkpoint_config_for_checklist(self) -> None:
        legacy_checkpoints: List[Union[Checkpoint, LegacyCheckpoint]] = []
        checkpoint: Union[Checkpoint, LegacyCheckpoint]
        checkpoint_name: str
        try:
            for checkpoint_name in sorted(self.data_context.list_checkpoints()):
                checkpoint = self.data_context.get_checkpoint(name=checkpoint_name)
                if checkpoint.config_version is None:
                    legacy_checkpoints.append(checkpoint)

            self.upgrade_checklist["manual"]["checkpoints"] = {
                checkpoint.name: checkpoint.get_config()
                for checkpoint in legacy_checkpoints
            }

            if len(self.upgrade_checklist["manual"]["checkpoints"]) == 0:
                self.upgrade_log["skipped_checkpoint_config_upgrade"] = True
        except gx_exceptions.InvalidTopLevelConfigKeyError:
            self.upgrade_log["skipped_checkpoint_config_upgrade"] = True

    # noinspection SpellCheckingInspection
    def _process_datasources_for_checklist(self) -> None:
        config_commented_map: CommentedMap = (
            self.data_context.get_config().commented_map
        )
        datasources: dict = config_commented_map.get("datasources") or {}
        datasource_name: str
        datasource_config: dict
        self.upgrade_checklist["manual"]["datasources"] = {
            datasource_name: datasource_config
            for datasource_name, datasource_config in datasources.items()
            if (
                set(datasource_config.keys())
                & {"execution_engine", "data_connectors", "introspection", "tables"}
            )
            == set()
        }

        if len(self.upgrade_checklist["manual"]["datasources"]) == 0:
            self.upgrade_log["skipped_datasources_upgrade"] = True

    def _process_validation_operators_for_checklist(self) -> None:
        config_commented_map: CommentedMap = (
            self.data_context.get_config().commented_map
        )
        validation_operators: dict = (
            config_commented_map.get("validation_operators") or {}
        )
        validation_operator_name: str
        validation_operator_config: dict
        self.upgrade_checklist["manual"]["validation_operators"] = {
            validation_operator_name: validation_operator_config
            for validation_operator_name, validation_operator_config in validation_operators.items()
            if validation_operator_config
        }

        if len(self.upgrade_checklist["manual"]["validation_operators"]) == 0:
            self.upgrade_log["skipped_validation_operators_upgrade"] = True

    def manual_steps_required(self) -> bool:
        return any(
            [
                len(manual_upgrade_item.keys()) > 0
                for manual_upgrade_item in self.upgrade_checklist["manual"].values()
            ]
        )

    def get_upgrade_overview(self) -> Tuple[str, bool]:
        manual_steps_required = self.manual_steps_required()

        increment_version = self.upgrade_log["update_version"]

        confirmation_required = (
            increment_version
            and not self.upgrade_log["skipped_checkpoint_store_upgrade"]
        )

        upgrade_overview = """\
<cyan>\
++====================================++
|| UpgradeHelperV13: Upgrade Overview ||
++====================================++\
</cyan>

"""
        if increment_version:
            upgrade_overview += (
                """\
UpgradeHelperV13 will upgrade your project to be compatible with Great Expectations V3 API.
"""
                + self._upgrade_overview_common_content(
                    manual_steps_required=manual_steps_required
                )
                + """
<cyan>\
Upgrade Confirmation
=====================
</cyan>
Please consult the V3 API migration guide for instructions on how to complete any required manual steps or to learn \
more about the automated upgrade process:

    <cyan>https://docs.greatexpectations.io/docs/guides/miscellaneous/migration_guide#migrating-to-the-batch-request-v3-api</cyan>
"""
            )
            if confirmation_required:
                upgrade_overview += """\

Would you like to proceed with the project upgrade?\
"""
        else:
            upgrade_overview += """\
Your project needs to be upgraded in order to be compatible with Great Expectations V3 API.
""" + self._upgrade_overview_common_content(
                manual_steps_required=manual_steps_required
            )

        return upgrade_overview, confirmation_required

    def _upgrade_overview_common_content(self, manual_steps_required: bool) -> str:
        stores_upgrade_checklist = list(
            self.upgrade_checklist["automatic"]["stores"].keys()
        )
        store_names_upgrade_checklist = list(
            self.upgrade_checklist["automatic"]["store_names"].keys()
        )

        checkpoint_config_upgrade_checklist = self.upgrade_checklist["manual"][
            "checkpoints"
        ]

        # noinspection SpellCheckingInspection
        datasources_upgrade_checklist = self.upgrade_checklist["manual"]["datasources"]

        upgrade_overview = ""

        # noinspection SpellCheckingInspection
        if (
            self.upgrade_log["skipped_checkpoint_store_upgrade"]
            and self.upgrade_log["skipped_checkpoint_config_upgrade"]
            and self.upgrade_log["skipped_datasources_upgrade"]
            and self.upgrade_log["skipped_validation_operators_upgrade"]
        ):
            upgrade_overview += """
<green>\
Good news! No special upgrade steps are required to bring your project up to date.
The Upgrade Helper will simply increment the config_version of your great_expectations.yml for you.
</green>
"""
        else:
            upgrade_overview += """
<red>**WARNING**: Before proceeding, please make sure you have appropriate backups of your project.</red>
"""
            if not self.upgrade_log["skipped_checkpoint_store_upgrade"]:
                if stores_upgrade_checklist or store_names_upgrade_checklist:
                    upgrade_overview += """
<cyan>\
Automated Steps
================
</cyan>
The following Stores and/or Store Names will be upgraded:

"""
                    upgrade_overview += (
                        f"""\
    - Stores: {", ".join(stores_upgrade_checklist)}
"""
                        if stores_upgrade_checklist
                        else ""
                    )
                    upgrade_overview += (
                        f"""\
    - Store Names: {", ".join(store_names_upgrade_checklist)}
"""
                        if store_names_upgrade_checklist
                        else ""
                    )

            if manual_steps_required:
                upgrade_overview += """
<cyan>\
Manual Steps
=============\
</cyan>
"""
                if not self.upgrade_log["skipped_checkpoint_config_upgrade"]:
                    upgrade_overview += """\

The following Checkpoints must be upgraded manually, due to using the old Checkpoint format, which is being deprecated:

"""
                    upgrade_overview += (
                        f"""\
    - Checkpoints: {", ".join(checkpoint_config_upgrade_checklist)}
"""
                        if checkpoint_config_upgrade_checklist
                        else ""
                    )

                # noinspection SpellCheckingInspection
                if not self.upgrade_log["skipped_datasources_upgrade"]:
                    upgrade_overview += """\

The following Data Sources must be upgraded manually, due to using the old Datasource format, which is being deprecated:

"""
                    upgrade_overview += (
                        f"""\
    - Data Sources: {", ".join(datasources_upgrade_checklist)}
"""
                        if datasources_upgrade_checklist
                        else ""
                    )

                if not self.upgrade_log["skipped_validation_operators_upgrade"]:
                    upgrade_overview += """
Your configuration uses validation_operators, which are being deprecated.  Please, manually convert \
validation_operators to use the new Checkpoint validation unit, since validation_operators will be deleted.

"""
            else:
                upgrade_overview += """
<cyan>\
Manual Steps
=============
</cyan>
No manual upgrade steps are required.
"""
        return upgrade_overview

    def upgrade_project(self):
        # noinspection PyBroadException
        try:
            self._upgrade_configuration_automatically()
        except Exception:
            pass

        # return a report of what happened, boolean indicating whether or not version should be incremented,
        # the report should include instructions for steps to be performed manually
        (
            upgrade_report,
            increment_version,
            exception_occurred,
        ) = self._generate_upgrade_report()
        return upgrade_report, increment_version, exception_occurred

    def _upgrade_configuration_automatically(self) -> None:
        if not self.upgrade_log["skipped_checkpoint_store_upgrade"]:
            config_commented_map: CommentedMap = (
                self.data_context.get_config().commented_map
            )
            for key, config in self.upgrade_checklist["automatic"]["stores"].items():
                config_commented_map["stores"][key] = config

            for key, value in self.upgrade_checklist["automatic"][
                "store_names"
            ].items():
                config_commented_map[key] = value

            data_context_config: DataContextConfig = (
                DataContextConfig.from_commented_map(commented_map=config_commented_map)
            )
            self.data_context.set_config(project_config=data_context_config)
            self.data_context._save_project_config()

            checkpoint_log_entry = {
                "stores": {
                    DataContextConfigDefaults.DEFAULT_CHECKPOINT_STORE_NAME.value: data_context_config.stores[
                        DataContextConfigDefaults.DEFAULT_CHECKPOINT_STORE_NAME.value
                    ],
                },
                "checkpoint_store_name": data_context_config.checkpoint_store_name,
            }
            self.upgrade_log["added_checkpoint_store"].update(checkpoint_log_entry)

    def _generate_upgrade_report(self):
        upgrade_log_path = self._save_upgrade_log()
        increment_version = self.upgrade_log["update_version"]
        upgrade_report = """\
<cyan>\
++================++
|| Upgrade Report ||
++================++\
</cyan>
"""
        manual_steps_required = self.manual_steps_required()

        if increment_version:
            if manual_steps_required:
                upgrade_report += f"""
<yellow>\
The Upgrade Helper has performed the automated upgrade steps as part of upgrading your project to be compatible with \
Great Expectations V3 API, and the config_version of your great_expectations.yml has been automatically incremented to \
3.0.  However, manual steps are required in order for the upgrade process to be completed successfully.

A log detailing the upgrade can be found here:

    - {upgrade_log_path}\
</yellow>\
"""
            else:
                upgrade_report += f"""
<green>\
Your project was successfully upgraded to be compatible with Great Expectations V3 API.  The config_version of your \
great_expectations.yml has been automatically incremented to 3.0.

A log detailing the upgrade can be found here:

    - {upgrade_log_path}\
</green>\
"""
        else:
            if manual_steps_required:
                upgrade_report += f"""
<yellow>\
The Upgrade Helper does not have any automated upgrade steps to perform as part of upgrading your project to be \
compatible with Great Expectations V3 API, and the config_version of your great_expectations.yml is already set to \
3.0.  However, manual steps are required in order for the upgrade process to be completed successfully.

A log detailing the upgrade can be found here:

    - {upgrade_log_path}\
</yellow>\
"""
            else:
                upgrade_report += f"""
<yellow>\
The Upgrade Helper finds your project to be compatible with Great Expectations V3 API, and the config_version of your \
great_expectations.yml is already set to 3.0.  There are no additional automatic or manual steps required, since the \
upgrade process has been completed successfully.

A log detailing the upgrade can be found here:

    - {upgrade_log_path}\
</yellow>\
"""
        exception_occurred = False
        return upgrade_report, increment_version, exception_occurred

    def _save_upgrade_log(self):
        current_time = datetime.datetime.now(datetime.timezone.utc).strftime(
            "%Y%m%dT%H%M%S.%fZ"
        )
        dest_path = os.path.join(
            self.data_context.root_directory,
            "uncommitted",
            "logs",
            "project_upgrades",
            f"UpgradeHelperV13_{current_time}.json",
        )
        dest_dir, dest_filename = os.path.split(dest_path)
        os.makedirs(dest_dir, exist_ok=True)

        with open(dest_path, "w") as outfile:
            json.dump(self.upgrade_log, outfile, indent=2)

        return dest_path



import copy
import json
import logging
import smtplib
import ssl
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import List, Optional, Union

import requests

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.batch import (
    BatchRequest,
    BatchRequestBase,
    RuntimeBatchRequest,
    get_batch_request_as_dict,
    materialize_batch_request,
)
from great_expectations.core.util import nested_update
from great_expectations.types import DictDot

try:
    import boto3
except ImportError:
    boto3 = None

logger = logging.getLogger(__name__)


def send_slack_notification(
    query, slack_webhook=None, slack_channel=None, slack_token=None
):
    session = requests.Session()
    url = slack_webhook
    query = query
    headers = None

    # Slack doc about overwritting the channel when using the legacy Incoming Webhooks
    # https://api.slack.com/legacy/custom-integrations/messaging/webhooks
    # ** Since it is legacy, it could be deprecated or removed in the future **
    if slack_channel:
        query["channel"] = slack_channel

    if not slack_webhook:
        url = "https://slack.com/api/chat.postMessage"
        headers = {"Authorization": f"Bearer {slack_token}"}

    try:
        response = session.post(url=url, headers=headers, json=query)
        if slack_webhook:
            ok_status = response.text == "ok"
        else:
            ok_status = response.json()["ok"]
    except requests.ConnectionError:
        logger.warning(f"Failed to connect to Slack webhook after {10} retries.")
    except Exception as e:
        logger.error(str(e))
    else:
        if response.status_code != 200 or not ok_status:
            logger.warning(
                "Request to Slack webhook "
                f"returned error {response.status_code}: {response.text}"
            )

        else:
            return "Slack notification succeeded."


# noinspection SpellCheckingInspection
def send_opsgenie_alert(query, suite_name, settings):
    """Creates an alert in Opsgenie."""
    if settings["region"] is not None:
        url = f"https://api.{settings['region']}.opsgenie.com/v2/alerts"  # accommodate for Europeans
    else:
        url = "https://api.opsgenie.com/v2/alerts"

    headers = {"Authorization": f"GenieKey {settings['api_key']}"}
    payload = {
        "message": f"Great Expectations suite {suite_name} failed",
        "description": query,
        "priority": settings["priority"],  # allow this to be modified in settings
        "tags": settings["tags"],
    }

    session = requests.Session()

    try:
        response = session.post(url, headers=headers, json=payload)
    except requests.ConnectionError:
        logger.warning("Failed to connect to Opsgenie")
    except Exception as e:
        logger.error(str(e))
    else:
        if response.status_code != 202:
            logger.warning(
                "Request to Opsgenie API "
                f"returned error {response.status_code}: {response.text}"
            )
        else:
            return "success"
    return "error"


def send_microsoft_teams_notifications(query, microsoft_teams_webhook):
    session = requests.Session()
    try:
        response = session.post(url=microsoft_teams_webhook, json=query)
    except requests.ConnectionError:
        logger.warning("Failed to connect to Microsoft Teams webhook after 10 retries.")

    except Exception as e:
        logger.error(str(e))
    else:
        if response.status_code != 200:
            logger.warning(
                "Request to Microsoft Teams webhook "
                f"returned error {response.status_code}: {response.text}"
            )
            return
        else:
            return "Microsoft Teams notification succeeded."


def send_webhook_notifications(query, webhook, target_platform):
    session = requests.Session()
    try:
        response = session.post(url=webhook, json=query)
    except requests.ConnectionError:
        logger.warning(
            f"Failed to connect to {target_platform} webhook after 10 retries."
        )
    except Exception as e:
        logger.error(str(e))
    else:
        if response.status_code != 200:
            logger.warning(
                f"Request to {target_platform} webhook "
                f"returned error {response.status_code}: {response.text}"
            )
        else:
            return f"{target_platform} notification succeeded."


# noinspection SpellCheckingInspection
def send_email(
    title,
    html,
    smtp_address,
    smtp_port,
    sender_login,
    sender_password,
    sender_alias,
    receiver_emails_list,
    use_tls,
    use_ssl,
):
    msg = MIMEMultipart()
    msg["From"] = sender_alias
    msg["To"] = ", ".join(receiver_emails_list)
    msg["Subject"] = title
    msg.attach(MIMEText(html, "html"))
    try:
        if use_ssl:
            if use_tls:
                logger.warning("Please choose between SSL or TLS, will default to SSL")
            context = ssl.create_default_context()
            mailserver = smtplib.SMTP_SSL(smtp_address, smtp_port, context=context)
        elif use_tls:
            mailserver = smtplib.SMTP(smtp_address, smtp_port)
            context = ssl.create_default_context()
            mailserver.starttls(context=context)
        else:
            logger.warning("Not using TLS or SSL to send an email is not secure")
            mailserver = smtplib.SMTP(smtp_address, smtp_port)
        mailserver.login(sender_login, sender_password)
        mailserver.sendmail(sender_alias, receiver_emails_list, msg.as_string())
        mailserver.quit()
    except smtplib.SMTPConnectError:
        logger.error(f"Failed to connect to the SMTP server at address: {smtp_address}")
    except smtplib.SMTPAuthenticationError:
        logger.error(
            f"Failed to authenticate to the SMTP server at address: {smtp_address}"
        )
    except Exception as e:
        logger.error(str(e))
    else:
        return "success"


def get_substituted_validation_dict(
    substituted_runtime_config: dict, validation_dict: dict
) -> dict:
    substituted_validation_dict = {
        "batch_request": get_substituted_batch_request(
            substituted_runtime_config=substituted_runtime_config,
            validation_batch_request=validation_dict.get("batch_request"),
        ),
        "expectation_suite_name": validation_dict.get("expectation_suite_name")
        or substituted_runtime_config.get("expectation_suite_name"),
        "expectation_suite_ge_cloud_id": validation_dict.get(
            "expectation_suite_ge_cloud_id"
        )
        or substituted_runtime_config.get("expectation_suite_ge_cloud_id"),
        "action_list": get_updated_action_list(
            base_action_list=substituted_runtime_config.get("action_list"),
            other_action_list=validation_dict.get("action_list", {}),
        ),
        "evaluation_parameters": nested_update(
            substituted_runtime_config.get("evaluation_parameters") or {},
            validation_dict.get("evaluation_parameters", {}),
            dedup=True,
        ),
        "runtime_configuration": nested_update(
            substituted_runtime_config.get("runtime_configuration") or {},
            validation_dict.get("runtime_configuration", {}),
            dedup=True,
        ),
        "include_rendered_content": validation_dict.get("include_rendered_content")
        or substituted_runtime_config.get("include_rendered_content")
        or None,
    }

    for attr in ("name", "id"):
        if validation_dict.get(attr) is not None:
            substituted_validation_dict[attr] = validation_dict[attr]

    validate_validation_dict(substituted_validation_dict)
    return substituted_validation_dict


# TODO: <Alex>A common utility function should be factored out from DataContext.get_batch_list() for any purpose.</Alex>
def get_substituted_batch_request(
    substituted_runtime_config: dict,
    validation_batch_request: Optional[Union[BatchRequestBase, dict]] = None,
) -> Optional[Union[BatchRequest, RuntimeBatchRequest]]:
    substituted_runtime_batch_request = substituted_runtime_config.get("batch_request")

    if substituted_runtime_batch_request is None and validation_batch_request is None:
        return None

    if substituted_runtime_batch_request is None:
        substituted_runtime_batch_request = {}

    if validation_batch_request is None:
        validation_batch_request = {}

    validation_batch_request = get_batch_request_as_dict(
        batch_request=validation_batch_request
    )
    substituted_runtime_batch_request = get_batch_request_as_dict(
        batch_request=substituted_runtime_batch_request
    )

    for key, value in validation_batch_request.items():
        substituted_value = substituted_runtime_batch_request.get(key)
        if value is not None and substituted_value is not None:
            raise gx_exceptions.CheckpointError(
                f'BatchRequest attribute "{key}" was specified in both validation and top-level CheckpointConfig.'
            )

    effective_batch_request: dict = dict(
        **substituted_runtime_batch_request, **validation_batch_request
    )

    return materialize_batch_request(batch_request=effective_batch_request)


def substitute_template_config(source_config: dict, template_config: dict) -> dict:
    if not (template_config and any(template_config.values())):
        return source_config

    dest_config: dict = copy.deepcopy(template_config)

    # replace
    if source_config.get("name") is not None:
        dest_config["name"] = source_config["name"]
    if source_config.get("module_name") is not None:
        dest_config["module_name"] = source_config["module_name"]
    if source_config.get("class_name") is not None:
        dest_config["class_name"] = source_config["class_name"]
    if source_config.get("run_name_template") is not None:
        dest_config["run_name_template"] = source_config["run_name_template"]
    if source_config.get("expectation_suite_name") is not None:
        dest_config["expectation_suite_name"] = source_config["expectation_suite_name"]
    if source_config.get("expectation_suite_ge_cloud_id") is not None:
        dest_config["expectation_suite_ge_cloud_id"] = source_config[
            "expectation_suite_ge_cloud_id"
        ]

    # update
    if source_config.get("batch_request") is not None:
        batch_request = dest_config.get("batch_request") or {}
        updated_batch_request = nested_update(
            batch_request,
            source_config["batch_request"],
            dedup=True,
        )
        dest_config["batch_request"] = updated_batch_request
    if source_config.get("action_list") is not None:
        action_list = dest_config.get("action_list") or []
        dest_config["action_list"] = get_updated_action_list(
            base_action_list=action_list,
            other_action_list=source_config["action_list"],
        )
    if source_config.get("evaluation_parameters") is not None:
        evaluation_parameters = dest_config.get("evaluation_parameters") or {}
        updated_evaluation_parameters = nested_update(
            evaluation_parameters,
            source_config["evaluation_parameters"],
            dedup=True,
        )
        dest_config["evaluation_parameters"] = updated_evaluation_parameters
    if source_config.get("runtime_configuration") is not None:
        runtime_configuration = dest_config.get("runtime_configuration") or {}
        updated_runtime_configuration = nested_update(
            runtime_configuration,
            source_config["runtime_configuration"],
            dedup=True,
        )
        dest_config["runtime_configuration"] = updated_runtime_configuration
    if source_config.get("validations") is not None:
        validations = dest_config.get("validations") or []
        existing_validations = template_config.get("validations") or []
        validations.extend(
            filter(
                lambda v: v not in existing_validations, source_config["validations"]
            )
        )
        dest_config["validations"] = validations
    if source_config.get("profilers") is not None:
        profilers = dest_config.get("profilers") or []
        existing_profilers = template_config.get("profilers") or []
        profilers.extend(
            filter(lambda v: v not in existing_profilers, source_config["profilers"])
        )
        dest_config["profilers"] = profilers

    return dest_config


def substitute_runtime_config(source_config: dict, runtime_kwargs: dict) -> dict:
    if not (runtime_kwargs and any(runtime_kwargs.values())):
        return source_config

    dest_config: dict = copy.deepcopy(source_config)

    # replace
    if runtime_kwargs.get("template_name") is not None:
        dest_config["template_name"] = runtime_kwargs["template_name"]
    if runtime_kwargs.get("run_name_template") is not None:
        dest_config["run_name_template"] = runtime_kwargs["run_name_template"]
    if runtime_kwargs.get("expectation_suite_name") is not None:
        dest_config["expectation_suite_name"] = runtime_kwargs["expectation_suite_name"]
    if runtime_kwargs.get("expectation_suite_ge_cloud_id") is not None:
        dest_config["expectation_suite_ge_cloud_id"] = runtime_kwargs[
            "expectation_suite_ge_cloud_id"
        ]
    # update
    if runtime_kwargs.get("batch_request") is not None:
        batch_request = dest_config.get("batch_request") or {}
        batch_request_from_runtime_kwargs = runtime_kwargs["batch_request"]
        batch_request_from_runtime_kwargs = get_batch_request_as_dict(
            batch_request=batch_request_from_runtime_kwargs
        )
        updated_batch_request = nested_update(
            batch_request,
            batch_request_from_runtime_kwargs,
            dedup=True,
        )
        dest_config["batch_request"] = updated_batch_request
    if runtime_kwargs.get("action_list") is not None:
        action_list = dest_config.get("action_list") or []
        dest_config["action_list"] = get_updated_action_list(
            base_action_list=action_list,
            other_action_list=runtime_kwargs["action_list"],
        )
    if runtime_kwargs.get("evaluation_parameters") is not None:
        evaluation_parameters = dest_config.get("evaluation_parameters") or {}
        updated_evaluation_parameters = nested_update(
            evaluation_parameters,
            runtime_kwargs["evaluation_parameters"],
            dedup=True,
        )
        dest_config["evaluation_parameters"] = updated_evaluation_parameters
    if runtime_kwargs.get("runtime_configuration") is not None:
        runtime_configuration = dest_config.get("runtime_configuration") or {}
        updated_runtime_configuration = nested_update(
            runtime_configuration,
            runtime_kwargs["runtime_configuration"],
            dedup=True,
        )
        dest_config["runtime_configuration"] = updated_runtime_configuration
    if runtime_kwargs.get("validations") is not None:
        validations = dest_config.get("validations") or []
        existing_validations = source_config.get("validations") or []
        validations.extend(
            filter(
                lambda v: v not in existing_validations,
                runtime_kwargs["validations"],
            )
        )
        dest_config["validations"] = validations
    if runtime_kwargs.get("profilers") is not None:
        profilers = dest_config.get("profilers") or []
        existing_profilers = source_config.get("profilers") or []
        profilers.extend(
            filter(lambda v: v not in existing_profilers, runtime_kwargs["profilers"])
        )
        dest_config["profilers"] = profilers

    return dest_config


def get_updated_action_list(
    base_action_list: list, other_action_list: list
) -> List[dict]:
    if base_action_list is None:
        base_action_list = []

    base_action_list_dict = {action["name"]: action for action in base_action_list}

    for other_action in other_action_list:
        other_action_name = other_action["name"]
        if other_action_name in base_action_list_dict:
            if other_action["action"]:
                nested_update(
                    base_action_list_dict[other_action_name],
                    other_action,
                    dedup=True,
                )
            else:
                base_action_list_dict.pop(other_action_name)
        else:
            base_action_list_dict[other_action_name] = other_action

    for other_action in other_action_list:
        other_action_name = other_action["name"]
        if other_action_name in base_action_list_dict:
            if not other_action["action"]:
                base_action_list_dict.pop(other_action_name)

    return list(base_action_list_dict.values())


def batch_request_in_validations_contains_batch_data(
    validations: Optional[List[dict]] = None,
) -> bool:
    if validations is not None:
        for idx, val in enumerate(validations):
            if (
                val.get("batch_request") is not None
                and isinstance(val.get("batch_request"), (dict, DictDot))
                and val["batch_request"].get("runtime_parameters") is not None
                and val["batch_request"]["runtime_parameters"].get("batch_data")
                is not None
            ):
                return True

    return False


def get_validations_with_batch_request_as_dict(
    validations: Optional[list] = None,
) -> Optional[list]:
    if validations:
        for value in validations:
            if "batch_request" in value:
                value["batch_request"] = get_batch_request_as_dict(
                    batch_request=value["batch_request"]
                )

    return validations


def validate_validation_dict(validation_dict: dict) -> None:
    if validation_dict.get("batch_request") is None:
        raise gx_exceptions.CheckpointError("validation batch_request cannot be None")
    if not validation_dict.get("expectation_suite_name"):
        raise gx_exceptions.CheckpointError(
            "validation expectation_suite_name must be specified"
        )
    if not validation_dict.get("action_list"):
        raise gx_exceptions.CheckpointError("validation action_list cannot be empty")


def send_sns_notification(
    sns_topic_arn: str, sns_subject: str, validation_results: str, **kwargs
) -> str:
    """
    Send JSON results to an SNS topic with a schema of:


    :param sns_topic_arn:  The SNS Arn to publish messages to
    :param sns_subject: : The SNS Message Subject - defaults to expectation_suite_identifier.expectation_suite_name
    :param validation_results:  The results of the validation ran
    :param kwargs:  Keyword arguments to pass to the boto3 Session
    :return:  Message ID that was published

    """
    if not boto3:
        logger.warning("boto3 is not installed")
        return "boto3 is not installed"

    message_dict = {
        "TopicArn": sns_topic_arn,
        "Subject": sns_subject,
        "Message": json.dumps(validation_results),
        "MessageAttributes": {
            "String": {"DataType": "String.Array", "StringValue": "ValidationResults"},
        },
        "MessageStructure": "json",
    }
    session = boto3.Session(**kwargs)
    sns = session.client("sns")
    try:
        response = sns.publish(**message_dict)
    except sns.exceptions.InvalidParameterException:
        logger.error(f"Received invalid for message: {validation_results}")
    else:
        return f"Successfully posted results to {response['MessageId']} with Subject {sns_subject}"



from typing import Dict

import pytest

from great_expectations.core.http import DEFAULT_TIMEOUT, create_session


@pytest.mark.unit
def test_session_factory_contains_appropriate_headers(
    ge_cloud_access_token: str, request_headers: Dict[str, str]
) -> None:
    session = create_session(access_token=ge_cloud_access_token)
    for key, val in request_headers.items():
        assert session.headers.get(key) == val


@pytest.mark.unit
def test_session_factory_contains_appropriate_adapters() -> None:
    access_token = "05f23701-d8e3-4c0a-8e38-2ad7bf16cb58"
    retry_count = 10
    backoff_factor = 0.1
    session = create_session(
        access_token=access_token,
        retry_count=retry_count,
        backoff_factor=backoff_factor,
    )

    assert len(session.adapters) == 2
    for adapter in session.adapters.values():
        assert adapter.timeout == DEFAULT_TIMEOUT
        retries = adapter.max_retries
        assert retries.total == retry_count
        assert retries.backoff_factor == backoff_factor



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

import click

from great_expectations.cli import toolkit
from great_expectations.cli.build_docs import build_docs
from great_expectations.cli.pretty_printing import cli_message, cli_message_list
from great_expectations.core.usage_statistics.events import UsageStatsEvents
from great_expectations.core.usage_statistics.util import send_usage_message
from great_expectations.exceptions import DataContextError

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.file_data_context import (
        FileDataContext,
    )


@click.group()
@click.pass_context
def docs(ctx: click.Context) -> None:
    """Data Docs operations"""
    ctx.obj.data_context = ctx.obj.get_data_context_from_config_file()

    invoked_subcommand = ctx.invoked_subcommand
    assert (
        invoked_subcommand
    ), "Proper registration of subcommand has not occurred; please review parent Click context"

    cli_event_noun: str = "docs"
    (
        begin_event_name,
        end_event_name,
    ) = UsageStatsEvents.get_cli_begin_and_end_event_names(
        noun=cli_event_noun,
        verb=invoked_subcommand,
    )
    send_usage_message(
        data_context=ctx.obj.data_context,
        event=begin_event_name,
        success=True,
    )
    ctx.obj.usage_event_end = end_event_name


@docs.command(name="build")
@click.option(
    "--site-name",
    "-sn",
    help="The site for which to generate documentation. If not present all sites will be built. See data_docs section in great_expectations.yml",
)
@click.option(
    "--no-view",
    "-nv",
    "no_view",
    is_flag=True,
    help="By default open in browser unless you specify the --no-view flag",
    default=False,
)
@click.pass_context
def docs_build(
    ctx: click.Context, site_name: Optional[str] = None, no_view: bool = False
) -> None:
    """Build Data Docs for a project."""
    context: FileDataContext = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end

    if site_name is not None and site_name not in context.get_site_names():
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message=f"<red>The specified site name `{site_name}` does not exist in this project.</red>",
        )
    if site_name is None:
        sites_to_build = context.get_site_names()
    else:
        sites_to_build = [site_name]

    build_docs(
        context,
        usage_stats_event=usage_event_end,
        site_names=sites_to_build,
        view=not no_view,
        assume_yes=ctx.obj.assume_yes,
    )
    send_usage_message(
        data_context=context,
        event=usage_event_end,
        success=True,
    )


@docs.command(name="list")
@click.pass_context
def docs_list(ctx: click.Context):
    """List known Data Docs sites."""
    context = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end
    docs_sites_url_dicts = context.get_docs_sites_urls()

    try:
        if len(docs_sites_url_dicts) == 0:
            cli_message("No Data Docs sites found")
        else:
            docs_sites_strings = [
                " - <cyan>{}</cyan>: {}".format(
                    docs_site_dict["site_name"],
                    docs_site_dict.get("site_url")
                    or f"site configured but does not exist. Run the following command to build site: great_expectations "
                    f'docs build --site-name {docs_site_dict["site_name"]}',
                )
                for docs_site_dict in docs_sites_url_dicts
            ]
            list_intro_string = _build_intro_string(docs_sites_strings)
            cli_message_list(docs_sites_strings, list_intro_string)

        send_usage_message(
            data_context=context,
            event=usage_event_end,
            success=True,
        )
    except Exception as e:
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message=f"<red>{e}</red>",
        )
        return


@docs.command(name="clean")
@click.option(
    "--site-name",
    "-s",
    help="The site that you want documentation cleaned for. See data_docs section in great_expectations.yml",
)
@click.option(
    "--all",
    "-a",
    "all_sites",
    is_flag=True,
    help="With this, all sites will get their data docs cleaned out. See data_docs section in great_expectations.yml",
)
@click.pass_context
def docs_clean(
    ctx: click.Context, site_name: Optional[str] = None, all_sites: bool = False
) -> None:
    """
    Remove all files from a Data Docs site.

    This is a useful first step if you wish to completely re-build a site from scratch.
    """
    context = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end

    if (site_name is None and all_sites is False) or (site_name and all_sites):
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message="<red>Please specify either --all to clean all sites or a specific site using --site-name</red>",
        )
    try:
        # if site_name is None, context.clean_data_docs(site_name=site_name)
        # will clean all sites.
        context.clean_data_docs(site_name=site_name)
        send_usage_message(
            data_context=context,
            event=usage_event_end,
            success=True,
        )
        cli_message("<green>Cleaned data docs</green>")
    except DataContextError as de:
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message=f"<red>{de}</red>",
        )


def _build_intro_string(docs_sites_strings: List[str]) -> str:
    doc_string_count = len(docs_sites_strings)
    if doc_string_count == 1:
        list_intro_string = "1 Data Docs site configured:"
    elif doc_string_count > 1:
        list_intro_string = f"{doc_string_count} Data Docs sites configured:"
    return list_intro_string
